#!/usr/bin/env python
# -*- coding: utf-8; py-indent-offset:4 -*-

# @Time    :2023/05/27 15:50:12
# @Author  :wakeblade (2390245#qq.com) 
# @version :8.1

"""
queue.py -- 任务处理器注册和任务注册的装饰器，以及启动任意多个Worker子进程
"""

from typing import Callable, Dict, List, Union
from multiprocessing import Process, Pool, cpu_count
from urllib import parse
from functools import partial
import psutil
import os
import signal
import platform
if platform.system()=='Windows':
    def oskill(pid:int, signal:signal):
        os.popen(f"taskkill /pid:{pid}")
else:
    def oskill(pid:int, signal:signal):
        os.kill(pid, signal)

from .client import Client, serialize, func2str
from .worker import Pusher, Worker, Assignor

def items2chunk(items:list, chunksize:int = 100):
    return (items[i*chunksize:(i+1)*chunksize] for i in range(len(items)//chunksize+1))

def items2dict(items:list):
    return {k:v for k,v in items}

def dict2chunk(d:dict, chunksize:int = 100):
    return [items2dict(items) for items in items2chunk(list(d.items()), chunksize)]

def start_worker(client_str:str, conn_url:str, assignor:Assignor = Assignor.PriorityOne, chunksize:int = 0, retrys:int = 10, retry_delay:int = 1):
    worker = Worker(client_str, conn_url, assignor, chunksize, retrys, retry_delay)
    worker.work()

def start_pusher(client_str:str, conn_url:str, topic:str = None, jobs:Union[List, Dict] = None):
    pusher = Pusher(client_str, conn_url)
    if topic is None:
        pusher.push_topics(jobs)
    else:
        pusher.push_topic(topic, jobs)

class FastQueue:

    _workers:Dict[str, Process] = {}

    def __init__(self, client:Client):
        self.client = client
        self._stop = False
        self.getJobs = []

    def register(self, topic:str, before:Callable = None, after:Callable = None):
        def decorator(handle:Callable):
            self.client.register(topic, handle, before, after)
            return handle
        return decorator        

    def jobs(self, topic:str = None):
        def decorator(func:Callable):
            self.getJobs.append(partial(func, topic))
            return self.getJobs
        return decorator        

    def pending(self, retrys:int = 1, retry_delay:float = 1.0, chunksize:int = 100, **kwargs):
        for getJob in self.getJobs:
            topic = getJob.args[0]
            if topic is None:
                for chunk in dict2chunk(jobs, chunksize):
                    for topic, data in chunk.items():
                        jobs = [serialize(d, retrys, retry_delay) for d in data]
                        self.client.push_topic(topic, jobs)
            else:
                jobs = [serialize(data, retrys, retry_delay) for data in getJob(**kwargs)]
                for chunk in items2chunk(jobs, chunksize):
                    self.client.push_topic(topic, chunk)

    #########################################################################################
    # 方案3.0，替换成Pool模式
    def pending_mp(self, client_str:str, conn_url:str, retrys:int = 1, retry_delay:float = 1.0, chunksize:int = 100, **kwargs):
        for getJob in self.getJobs:
            topic = getJob.args[0]
            if topic is None:
                jobs = {topic:serialize(data, retrys, retry_delay) for topic, data in getJob(**kwargs).items()}
                chunks = dict2chunk(jobs, chunksize)
            else:
                jobs = [serialize(data, retrys, retry_delay) for data in getJob(**kwargs)]
                chunks = items2chunk(jobs, chunksize)
            with Pool(processes=cpu_count()) as pool:
                for chunk in chunks:
                    pool.apply_async(
                        start_pusher,
                        (client_str, conn_url, topic, chunk)
                    )
                pool.close()
                pool.join()

    #########################################################################################
    # 方案2.0，可以任务数据函数用装饰器注入self.getJobs
    # def start_pusher(self, client_str:str, conn_url:str, topic:str = None, jobs:Union[List, Dict] = None, daemon:bool = True):
    #     process = Process(
    #         target=start_pusher, 
    #         args=(client_str, conn_url, topic, jobs))
    #     process.daemon = daemon
    #     process.start()
    #     print("#"*80)
    #     return process

    # def pending_mp(self, client_str:str, conn_url:str, retrys:int = 1, retry_delay:float = 1.0, chunksize:int = 100):
    #     if self.getJobs is not None:
    #         topic = self.getJobs.args[0]
    #         if topic is None:
    #             jobs = {topic:serialize(data, retrys, retry_delay) for topic, data in self.getJobs().items()}
    #             for chunk in dict2chunk(jobs, chunksize):
    #                 process = self.start_pusher(client_str, conn_url, None, chunk, daemon=False)
    #                 process.join()
    #         else:
    #             jobs = [serialize(data, retrys, retry_delay) for data in self.getJobs()]
    #             for chunk in items2chunk(jobs):
    #                 process = self.start_pusher(client_str, conn_url, topic, chunk, daemon=False)
    #                 process.join()

    #########################################################################################
    # 方案1.0，只能把任务数据作为参数注入
    # def topic_mp(self, topic:str, datas:list, retrys:int = 1, retry_delay:float = 1.0, chunksize:int = 100):
    #     client_str = func2str(self.client)
    #     conn_url = self.client.conn_params.geturl()
    #     jobs = [serialize(data, retrys, retry_delay) for data in datas]
    #     for chunk in items2chunk(jobs):
    #         process = self.start_pusher(client_str, conn_url, topic, chunk, daemon=False)
    #         process.join()

    # def topics_mp(self, datas:Dict[str, List], retrys:int = 1, retry_delay:float = 1.0, chunksize:int = 1):
    #     client_str = func2str(self.client)
    #     conn_url = self.client.conn_params.geturl()
    #     d = {topic:serialize(data, retrys, retry_delay) for topic, data in datas.items()}
    #     for chunk in dict2chunk(d, chunksize):
    #         process = self.start_pusher(client_str, conn_url, None, chunk, daemon=False)
    #         process.join()

    def topic(self, topic:str, retrys:int = 1, retry_delay:float = 1.0, chunksize:int = 100):
        def decorator(func:Callable):
            jobs = [serialize(data, retrys, retry_delay) for data in func()]
            for chunk in items2chunk(jobs, chunksize):
                self.client.push_topic(topic, chunk)
        return decorator

    def topics(self, retrys:int = 1, retry_delay:float = 1.0, chunksize:int = 100):
        def decorator(func:Callable):
            for chunk in dict2chunk(func(), chunksize):
                for topic, data in chunk.items():
                    jobs = [serialize(d, retrys, retry_delay) for d in data]
                    self.client.push_topic(topic, jobs)
        return decorator

    def clear_worker(self):
        # 收集死进程
        _tobekill = []
        for pid, process in self._workers.items():
            if not process.is_alive():
                _tobekill.append(pid)

        if len(self._workers)==len(_tobekill):
            self._stop = True

        # 清理死进程
        for pid in _tobekill:
            self._workers.pop(pid)
            # os.kill(pid, signal.CTRL_C_EVENT)
            if psutil.pid_exists(pid):
                # print(f"Workser({pid}): 销毁成功！", "#"*40)
                oskill(pid, signal.SIGBREAK)
        
        return len(_tobekill) + len(self._workers)
        
    def start_worker(self, client_str:str, conn_url:str, assignor:Assignor = Assignor.PriorityOne, chunksize:int = 100, retrys:int = 10, retry_delay:int = 1, daemon:bool = True):
        process = Process(
            target=start_worker, 
            args=(client_str, conn_url, assignor, chunksize, retrys, retry_delay))
        process.daemon = daemon
        process.start()
        return process

    def start_workers(self, client_str:str, conn_url:str, workers:int = 1, assignor:Assignor = Assignor.PriorityOne, chunksize:int = 100, retrys:int = 10, retry_delay:int = 1):
        # 补充新进程
        while not self._stop:
            while(workers-1>len(self._workers)):
                process = self.start_worker(client_str, conn_url, assignor, chunksize, retrys, retry_delay)
                self._workers[process.pid] = process
                # print(f"Workser({process.pid}): 创建成功({workers}/{len(self._workers)})！", "#"*40)
            workers = self.clear_worker()
        # for pid, process in self._workers.items():
        #     process.close()
        #     process.join()

    def start(self, workers:int = 1, assignor:Assignor = Assignor.PriorityOne, chunksize:int = 100, retrys:int = 10, retry_delay:int = 1, **kwargs):
        client_str = func2str(self.client)
        self.pending(retrys, retry_delay, **kwargs)
        self.start_workers(client_str, self.client.conn_url, workers, assignor, chunksize, retrys, retry_delay)
        
    def start_mp(self, workers:int = 1, assignor:Assignor = Assignor.PriorityOne, chunksize:int = 100, retrys:int = 10, retry_delay:int = 1, **kwargs):
        client_str = func2str(self.client)
        self.pending_mp(client_str, self.client.conn_url, retrys, retry_delay, chunksize, **kwargs)
        self.start_workers(client_str, self.client.conn_url, workers, assignor, chunksize, retrys, retry_delay)



#!/usr/bin/env python
# -*- coding: utf-8; py-indent-offset:4 -*-

# @Time    :2023/05/27 16:23:29
# @Author  :wakeblade (2390245#qq.com) 
# @version :8.1

"""
worker.py -- 任务处理子进程
"""

from enum import Enum
import time
from collections import defaultdict
from functools import reduce
import os

from .client import Client, str2func, unserialize

class Assignor(Enum):
    PriorityOne = 0 # 按优先级广度遍历
    PriorityAll = 1 # 按优先级深度遍历
    RoundRobinOne = 2 # 轮询广度遍历
    RoundRobinAll = 3 # 轮询深度遍历

class Pusher:
    
    def __init__(self, client_str:str, conn_url:str):
        client_class = str2func(client_str)
        self.client:Client = client_class(conn_url)

    def push_topic(self, topic:str, jobs:list):
        # print(f"Pusher({os.getpid()}): {len(jobs)}", "#"*40)
        return self.client.push_topic(topic, jobs)

    def push_topics(self, jobs:dict):
        # print(f"Pusher({os.getpid()}): {len(jobs)}", "#"*40)
        return self.client.push_topics(jobs)

class Worker:

    _handlers = {}
    _jobs = defaultdict(list)
    _stop = False 
    _topic = None

    def __init__(self, client_str:str, conn_url:str, assignor:Assignor = Assignor.PriorityOne, chunksize:int = 1, retrys:int = 10, retry_delay:int = 1):
        client_class = str2func(client_str)
        self.client:Client = client_class(conn_url)
        self.assignor = assignor
        self.chunksize = chunksize 
        self.retrys = retrys
        self.retry_delay = retry_delay

    def len(self):
        if len(self._jobs)<1:
            return 0
        
        if len(self._jobs)<2:
            return [len(self._jobs[k]) for k in self._jobs][0]

        return reduce(lambda x,y:x+y, [len(self._jobs[k]) for k in self._jobs])

    def load_handlers(self, topic:str):
        topic = topic if topic.find(b"#")<0 else topic[:topic.find(b"#")]
        if topic not in self._handlers:
            handlers = eval(self.client.get_handlers(topic))
            self._handlers[topic] = tuple(str2func(handle) for handle in handlers)
        return self._handlers[topic]

    def get_jobs(self, topic:str):
        retry_times = 0
        while retry_times<self.retrys:
            jobs = self.client.get_jobs(topic, self.chunksize)
            if jobs is not None and len(jobs)>0:
                return jobs
            retry_times+=1
            time.sleep(self.retry_delay)
        return None

    def get_job(self, topic:str):
        retry_times = 0
        while retry_times<self.retrys:
            job = self.client.get_job(topic)
            if job is not None and len(job)>0:
                return job
            retry_times+=1
            time.sleep(self.retry_delay)
        return None

    def load_jobs(self):
        topics = self.client.get_topics()
        topics_reported = self.client.get_topics_reported()

        last_priority = priority = -1
        while self.len()<self.chunksize: 
            last_len = self.len()
            if self._topic:
                jobs = self.get_jobs(self._topic)
                if jobs is not None:
                    for job in jobs:
                        self._jobs[topic].append(unserialize(job))
            else:
                for topic in topics:
                    if topic not in topics_reported and topic[-1]=="@":
                        self.client.report(topic, os.getpid())
                        self._topic = topic
                        break

                    if (self.assignor.value%2)==1:
                        jobs = self.get_jobs(topic)
                        if jobs is not None:
                            for job in jobs:
                                self._jobs[topic].append(unserialize(job))
                    else:
                        job = self.get_job(topic)
                        if job is not None:
                            self._jobs[topic].append(unserialize(job))
                    
                    if last_priority==priority or last_priority<1:
                        continue
            
            if last_len == self.len():
                break

        return self._jobs

    def work(self):
        _retry_times = 0
        while not self._stop and _retry_times<self.retrys:
            _jobs = self.load_jobs()
            if len(_jobs)<1:
                break

            topic, jobs = _jobs.popitem()
            if len(jobs)<1:
                time.sleep(self.retry_delay)
                _retry_times +=1
                # print(f"Workser({os.getpid()}): _retry_times = {_retry_times}", "#"*40)
            else:
                _retry_times = 0

            handle, before, after = self.load_handlers(topic)

            context = before(topic) if before else {"_result":[], "client":self.client}
            for job in jobs:
                data, retrys, retry_delay = job["data"], int(job["retrys"]), float(job["retry_delay"])
                retry_times = 0
                while retry_times<retrys:
                    try:
                        rs = handle(data, context)
                        context["_result"].append(rs)
                        break
                    except:
                        retry_times +=1
                        time.sleep(retry_delay)
                        # print(f"Job({os.getpid()}): _retry_times = {_retry_times}", "#"*40)
                        continue
             
            if after is not None:
                after(data, context)
            elif "topic" in context:
                self.client.push_topic(context["topic"], context["_result"])
            elif "topics" in context:
                self.client.push_topics(context["topics"])
            # print(f"Workser({os.getpid()}): {len(context['_result'])}", "#"*40)
        
        if self._topic:
            self.client.unreport(topic)
        


#!/usr/bin/env python
# -*- coding: utf-8; py-indent-offset:4 -*-

# @Time    :2023/05/27 12:14:34
# @Author  :wakeblade (2390245#qq.com) 
# @version :8.1

"""
client.py -- 访问保存任务主题、任务处理器和任务队列的消息队列的客户端
"""

from abc import ABC, abstractmethod
from typing import Type, Callable, List, Dict, Union, Any
from urllib import parse
from redis import Redis, ConnectionPool
import importlib
import json

def func2str(func:Union[Callable, Type]):
    if callable(func):
        return f"{func.__module__}.{func.__name__}"
    
    return f"{func.__module__}.{func.__class__.__name__}"

def str2func(s:str):
    if len(s)<1:
        return None
    
    m, _, f = s.rpartition(".")
    module = importlib.import_module(m)
    return getattr(module, f)

def serialize(data:Any, retrys:int = 1, retry_delay:float = 0.01):
    d = dict(
        data=data if isinstance(data, str) else json.dumps(data),
        retrys=retrys,
        retry_delay=retry_delay
    )
    return json.dumps(d)

def unserialize(data:str):
    s = str(data, encoding="utf8")
    return json.loads(s) if s[0] in "{[" else s

class Client(ABC):

    default_topics_header = "fasttq:topics"
    default_workers_header = "fasttq:workers"
    default_handlers_header = "fasttq:handlers"
    default_jobs_header = "fasttq:jobs:%s"

    conn_url:str = None

    @abstractmethod
    def connect(self):
        pass

    @abstractmethod
    def register(self, topic:str, handle:Callable, before:Callable, after:Callable, priority:int = 0):
        pass

    @abstractmethod
    def unregister(self, topic:str):
        pass

    @abstractmethod
    def push_topic(self, topic:str, jobs:List[str]):
        pass

    @abstractmethod
    def push_topics(self, jobs:Dict[str, str], retrys:int = 1, retry_delay:float = 1.0):
        pass

    @abstractmethod
    def insert_topic(self, topic:str, jobs:List[str]):
        pass

    @abstractmethod
    def insert_topics(self, jobs:Dict[str, str]):
        pass

    @abstractmethod
    def get_topics(self):
        pass

    @abstractmethod
    def get_handlers(self, topic:str):
        pass

    @abstractmethod
    def get_job(self, topic:str):
        pass

    @abstractmethod
    def get_jobs(self, topic:str, chunksize:int = 1):
        pass

class RedisClient(Client):

    def __init__(self, conn_url:str):
        self.conn_url = conn_url
        self.pool = ConnectionPool.from_url(conn_url)

    def connect(self):
        return Redis(connection_pool=self.pool)

    # 注册Topic以及处理器
    def register(self, topic:str, handle:Callable, before:Callable, after:Callable, priority:int = 0):
        with self.connect() as conn:
            t = conn.zadd(self.default_topics_header, {topic:priority,})
            handle_str = func2str(handle)
            before_str = func2str(before) if before else ""
            after_str = func2str(after) if after else ""
            h = conn.hset(self.default_handlers_header, key=topic, value=f"('{handle_str}', '{before_str}', '{after_str}')")
            return t,h

    # 注销Topic以及处理器
    def unregister(self, topic:str):
        with self.connect() as conn:
            w = conn.zrem(self.default_workers_header, topic)
            t = conn.zrem(self.default_topics_header, topic)
            h = conn.hdel(self.default_handlers_header, topic)
            j = conn.delete(self.default_jobs_header % topic)
            return w,t,h,j

    # 推入某个topic的任务
    def push_topic(self, topic:str, jobs:List[str]):
        with self.connect() as conn:
            return conn.lpush(self.default_jobs_header % topic, *jobs)

    # 推入多个topic的任务
    def push_topics(self, jobs:Dict[str, List], retrys:int = 1, retry_delay:float = 1.0):
        with self.connect() as conn:
            return {topic:conn.lpush(self.default_jobs_header % topic, *_jobs) for topic,_jobs in jobs.items()}

    # 优先插入某个topic的任务
    def insert_topic(self, topic:str, jobs:List[str]):
        with self.connect() as conn:
            return conn.rpush(self.default_jobs_header % topic, *jobs)

    # 优先插入多个topic的任务
    def insert_topics(self, jobs:Dict[str, str]):
        with self.connect() as conn:
            return {topic:conn.rpush(self.default_jobs_header % topic, *_jobs) for topic,_jobs in jobs.items()}
            
    # 获取所有活跃的topic
    def get_topics(self):
        with self.connect() as conn:
            return (key[len(self.default_jobs_header)-2:] for key in conn.keys(pattern=self.default_jobs_header % "*"))
            # count = conn.zcard(self.default_topics_header)
            # return conn.zrange(self.default_topics_header, 0 , count, withscores=True).extend(topics)

    # 获取某个topic的处理器
    def get_handlers(self, topic:str):
        with self.connect() as conn:
            return conn.hget(self.default_handlers_header, topic)

    # 获取某个topic的待处理任务
    def get_job(self, topic:str):
        with self.connect() as conn:
            return conn.rpop(self.default_jobs_header % unserialize(topic))

    # 获取某个topic的待处理任务
    def get_jobs(self, topic:str, chunksize:int = 10):
        with self.connect() as conn:
            header = self.default_jobs_header % unserialize(topic)
            return [conn.rpop(header) for _ in range(chunksize)]

    # worker报到
    def report(self, topic:str, worker:str):
        with self.connect() as conn:
            return conn.hset(self.default_workers_header, key=topic, value=worker)

    # worker告退
    def unreport(self, topic:str):
        with self.connect() as conn:
            return conn.hdel(self.default_workers_header, key=topic)

    # 获取有报到的所有topic
    def get_topics_reported(self):
        with self.connect() as conn:
            return conn.hkeys(self.default_workers_header)



"""
FastQ是一款由 [德波量化](http://www.dealbot.cn) 开源的基于消息队列的多进程分布式任务调度器，原应用于Dealbot量化策略高并发回测。
目前版本暂时仅支持Python语言，消息队列仅支持Redis。
"""
import os
from setuptools import find_packages, setup


def get_requirements():
    basedir = os.path.dirname(__file__)
    try:
        with open(os.path.join(basedir, 'requirements.txt')) as f:
            return f.readlines()
    except FileNotFoundError:
        raise RuntimeError('No requirements info found.')


setup(
    name='fasttq',
    version=__import__('fasttq').__version__,
    url='https://github.com/wakeblade/fasttq',
    license='BSD',
    author='Wakeblade',
    author_email='2390245@qq.com',
    description='FastTQ是一款由 [德波量化](http://www.dealbot.cn) 开源的基于消息队列的多进程分布式任务调度器',
    long_description=__doc__,
    packages=find_packages(exclude=['examples', '*.py']),
    package_data={"fasttq": ["py.typed"]},
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=get_requirements(),
    python_requires='>=3.6',
    entry_points={},
    classifiers=[
        # As from http://pypi.python.org/pypi?%3Aaction=list_classifiers
        # 'Development Status :: 1 - Planning',
        # 'Development Status :: 2 - Pre-Alpha',
        # 'Development Status :: 3 - Alpha',
        # 'Development Status :: 4 - Beta',
        'Development Status :: 5 - Production/Stable',
        # 'Development Status :: 6 - Mature',
        # 'Development Status :: 7 - Inactive',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX',
        'Operating System :: MacOS',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Internet',
        'Topic :: Scientific/Engineering',
        'Topic :: System :: Distributed Computing',
        'Topic :: System :: Systems Administration',
        'Topic :: System :: Monitoring',
    ],
)
